#!/usr/bin/python
#
# VERSION = 1.21 (08.02.2009)
#
# Changelog:
#     1.0 
#            - single oid request support
#     1.1 
#            - multiple mib request support (but no single oids) - api not changed
#     1.2 
#            - multible mib and multible oid support (api changed)
#     1.21
#            - the getOid function can now return an mib-subtree result like the getMib function


from pysnmp.carrier.asynsock.dispatch   import AsynsockDispatcher
from pysnmp.carrier.asynsock.dgram      import udp
from pysnmp.proto                       import api
from pysnmp.entity.rfc3413.oneliner     import cmdgen
from pyasn1.codec.ber                   import encoder, decoder

from pprint import pprint
from time   import time

import socket

class Manager:
    def __init__(self, agentHost, agentPort, agentCommunity, agentName='my-agent', timeout=60):
        self.agentHost        = agentHost
        self.agentPort        = int(agentPort)
        self.agentCommunity   = agentCommunity
        
        self.agentName        = agentName
        self.timeout          = timeout
        
        # mib-request command gen
        self.commandGenerator = cmdgen.CommandGenerator()
        
        # oid-request transport dispatcher
        self.protocolModule = api.protoModules[api.protoVersion1]

        self.transportDispatcher = AsynsockDispatcher()
        self.transportDispatcher.registerTransport(
            udp.domainName,
            udp.UdpSocketTransport().openClientMode()
        )

        self.transportDispatcher.registerRecvCbFun(self.handleRecieve)
        self.transportDispatcher.registerTimerCbFun(self.handleTimeout)

        self.result = {}


    def getMib(self, mibs):
        """
        Mib must be an iterable object containing tuples; e.g. (1,3,6,1,2,1,1,1,0)
        
        If a requested mib should be an oid the result will be None.
        (e.g.)  MIB - (1,3,6,1,2,1,1,1)   - returns values of all oids in mib (1,3,6,1,2,1,1,1)
                OID - (1,3,6,1,2,1,1,1,0) - this will return None
        Use the getOid function to get single oids.
        
        @type oids: iterable
        
        Returns a dictionary of all selected mibs (tuples) as keys and their containing oids<dict> as value.
        Every oid-dict contains the values as subelements.
        { mib : {oid1 : value1,
                 oid2 : value2,
                 ...
                 }
        }
        
        ##############
        
        { (1,3,6,1,2,1,1,1) : {
                                (1,3,6,1,2,1,1,1,0) : "Cisco xyz ..." ,
                              }
        }
        """
        
        # make tuple object
        mibs = tuple(mibs)
        #print oids
        
        # {mib : {oid : value}}
        self.result = {}
        
        for mib in mibs:
            self.result[mib] = {}
        
        # {oid : value}
        self.converted = {}
        
        try:
            errorIndication, errorStatus, errorIndex, varBinds = cmdgen.CommandGenerator().nextCmd(
                cmdgen.CommunityData(self.agentName, self.agentCommunity, 0),
                # snmp3 only
                #cmdgen.UsmUserData('my-user', 'my-authkey', 'my-privkey'),
                cmdgen.UdpTransportTarget((self.agentHost, self.agentPort)),
                *mibs
                #*((1, 3, 6, 1, 2, 1, 1, 1),)
            )
        except socket.gaierror, e:
            print "== SNMP ERROR =="
            print "error: could not resolve hostname: %s" % self.agentHost
            raise SNMPError
            
        if errorIndication is not None:
            print "== SNMP ERROR =="
            print "errorIndication: " + str(errorIndication)
            print "errorStatus:     " + str(errorStatus)
            print "errorIndex:      " + str(errorIndex)
            raise SNMPError
            
        if varBinds == []:
            return None
        
        # varBinds is a big list of lists
        # if you want to ask for more than one mib-subtree the result looks like follows
        # > get(oid1, oid2)
        # [[mib1-value1, mib2-value2], [mib1-value2, mib2-value2], [mib1-value3]]
        #
        # the amount of results of every mib-subtree is not konstant
        for item in varBinds:
            # item: [mib1-value, mib2-value]
            for subItem in item:
                # subitem: (oid, value)
                
                oid   = subItem[0]
                value = subItem[1]
                
                self.converted[oid] = value
                
        # result is a mapping
        # {oid : value, ...}
        
        for oid in self.converted.keys():
            value = self.converted[oid]
            
            for requestedMib in mibs:
                
                #print oid[:len(requestedOid) ], requestedOid
                
                if oid[:len(requestedMib)] == requestedMib:
                    self.result[requestedMib][tuple(oid)] = value
                    break
            
        return self.result
                
        #print "errorIndication:", errorIndication
        #print "errorStatus:    ", errorStatus
        #pprint(varBinds[0])        

    def getOid(self, oids, mibSubtree=True):
        """
        oids must be an iterable object containing tuples; e.g. (1,3,6,1,2,1,1,1,0)

        @type oids: iterable

        returns a dictionary of all selected oids (tuples) as keys and their value
        { (1,3,6,1,2,1,1,1,0) : "Cisco xyz ..." }
        
        if mibSubtree is given the output looks like follows:
        { mib: {oid : value, ...},
          ...
        }
        """
        self.mibSubtree = mibSubtree
        
         # reset the last result
        self.result = {}

        # is a list containing tuples consisting of 2 items, the oid as a tuple and a protoModule.Null instance
        # e.g.:
        # [ ((1,3,6,1,2,1,1,1,0), protocolModule.Null('')), (... ]
        requestedOids = []

        for oid in oids:
            # fill requestedOids
            requestedOids.append((oid, self.protocolModule.Null('')))

        # convert the requestedOids list to tuple
        requestedOids = tuple(requestedOids)

        self.requestPDU = self.protocolModule.GetRequestPDU()
        self.protocolModule.apiPDU.setDefaults(self.requestPDU)

        self.protocolModule.apiPDU.setVarBinds(self.requestPDU, requestedOids)

        # Build message
        requestMessage = self.protocolModule.Message()
        self.protocolModule.apiMessage.setDefaults(requestMessage)
        self.protocolModule.apiMessage.setCommunity(requestMessage, self.agentCommunity)
        self.protocolModule.apiMessage.setPDU(requestMessage, self.requestPDU)

        self.transportDispatcher.sendMessage(
            encoder.encode(requestMessage),
            udp.domainName,
            (self.agentHost, self.agentPort)
        )

        self.transportDispatcher.jobStarted(1)
        self.transportDispatcher.runDispatcher()

        return self.result

    def handleTimeout(self, timeNow, startedAt=time()):
        if timeNow - startedAt > self.timeout:
            raise Timeout

    def handleRecieve(self, transportDispatcher, transportDomain, transportAddress, wholeMessage, requestPDU=None):
        if not requestPDU:
            requestPDU = self.requestPDU

        while wholeMessage:
            responseMessage, wholeMessage = decoder.decode(wholeMessage, asn1Spec=self.protocolModule.Message())
            responsePDU = self.protocolModule.apiMessage.getPDU(responseMessage)

            # Match response to request
            if self.protocolModule.apiPDU.getRequestID(requestPDU) == self.protocolModule.apiPDU.getRequestID(responsePDU):
                # Check for SNMP errors reported
                self.errorStatus = self.protocolModule.apiPDU.getErrorStatus(responsePDU)
                if self.errorStatus:
                    print self.errorStatus.prettyPrint()
                else:
                    for oid, val in self.protocolModule.apiPDU.getVarBinds(responsePDU):
                        # adding to result dict
                        # oid._value -> (1,3,6,1,2,1,1,1,0) # oid as tuple
                        # val._value -> result as string
                        if self.mibSubtree:
                            mib = oid._value[:-1]
                            if not self.result.has_key(mib):
                                self.result[mib] = {}
                            self.result[mib][oid._value] = val._value
                        else:
                            self.result[oid._value] = val._value
                        
                        # debug it
                        #print '%s = %s' % (oid.prettyPrint(), val.prettyPrint())

                    transportDispatcher.jobFinished(1)

        return wholeMessage
    
class Timeout(Exception):
    """
    """
    
class SNMPError(Exception):
    """
    """
    
def convertOid(oid):
    """ 
    oid is a string like '.1.3.6.1.2.1.1.3.0'
    
    returns oid as tuple
    """
    
    # split the imput string
    splittedOid = oid.split(".")
    
    # delete first item if empty
    if splittedOid[0] == "":
        del splittedOid[0]

    # convert (map) the content of splitted oid to int
    splittedOid = map(lambda x: int(x), splittedOid)
    
    # convert the list to int
    oidTuple = tuple(splittedOid) 
    
    return oidTuple
    
if __name__ == "__main__":
    manager = Manager(agentHost="10.1.0.254",           agentPort=161, agentCommunity="INQBUS")
    #manager = Manager(agentHost="nrvoice12.aida.de",    agentPort=161, agentCommunity="inqbus")
                      
    from pprint import pprint
    
    # device description, uptime
    #result = manager.getOid([(1,3,6,1,2,1,1,1,0),])
    result = manager.getMib([(1,3,6,1,2,1,1,1),])
    
    # temperature (not supported at inqbus) - should be None (not supported)
    #result = manager.getObject([(1,3,6,1,4,1,9,9,13,1,3,1,6,1),])
     
    # memory usage
    #result = manager.getObject([convertOid(".1.3.6.1.4.1.9.9.48.1.1.1.5.1")])
    
    # cpuload 5min, 1min
    #result = manager.getObject(((1,3,6,1,2,1,1,1),))
    #result = manager.getObject(((1,3,6,1,2,1,1,1), (1,3,6,1,2,1,2,2,1,3)))
    
    # ifNumber.0
#    result = manager.getObject([convertOid(".1.3.6.1.2.1.2.2.1.3.10001")])
    pprint(result)
    